from ._version import __version__  # noqa: F401
from .check import check_bib_database  # noqa: F401
from .clean import clean_bib_database  # noqa: F401
from .utils.config import sys_info  # noqa: F401
from .utils.logs import add_file_handler, logger, set_log_level  # noqa: F401



from pathlib import Path
from typing import Union

from bibtexparser import dump, load
from bibtexparser.bibdatabase import BibDatabase

from .utils._checks import check_type
from .utils._docs import fill_doc
from .utils.logs import logger


@fill_doc
def load_bib(file: Union[str, Path], encoding: str = "utf-8") -> BibDatabase:
    """Load a BibTex file.

    Parameters
    ----------
    file : str | Path
        Path to the ``.bib`` file to load.
    %(encoding)s

    Returns
    -------
    bib_database : ``BibDatabase``
        BibTex database loaded.
    """
    check_type(file, (str, Path), "file")
    file = Path(file) if isinstance(file, str) else file
    if file.suffix != ".bib":
        raise IOError(
            f"The provided file extension is not '.bib'. '{file.suffix}' is invalid."
        )
    if not file.exists():
        raise IOError("The provided file does not exist.")

    logger.info("Loading file %s", file)
    with open(file, "r", encoding=encoding) as bibtex_file:
        bib_database = load(bibtex_file)
    return bib_database


@fill_doc
def save_bib(
    bib_database: BibDatabase,
    file: Union[str, Path],
    encoding: str = "utf-8",
    overwrite: bool = False,
) -> None:
    """Save a BibTex file.

    Parameters
    ----------
    bib_database : ``BibDatabase``
        BibTex database to save.
    file : str | Path
        Path to the ``.bib`` file to save.
    %(encoding)s
    overwrite : bool
        If True, an existing file will be overwritten.
    """
    check_type(bib_database, (BibDatabase,), "bib_database")
    check_type(file, (str, Path), "file")
    file = Path(file) if isinstance(file, str) else file
    if file.suffix != ".bib":
        raise IOError(
            f"The provided file extension is not '.bib'. '{file.suffix}' is invalid."
        )
    check_type(overwrite, (bool,), "overwrite")
    if file.exists() and not overwrite:
        raise IOError(
            "The provided file already exist. Set overwrite to True if you "
            "want to overwrite the file."
        )

    logger.info("Saving to file %s", file)
    with open(file, "w", encoding=encoding) as bibtex_file:
        dump(bib_database, bibtex_file)



from typing import Dict, List, Optional, Set

import numpy as np
from bibtexparser.bibdatabase import BibDatabase

from ._exception import DuplicateEntry, MissingReqField
from ._typing import Entry
from .config import _load_default_config
from .utils._checks import check_type, check_value
from .utils._docs import fill_doc


@fill_doc
def check_bib_database(
    bib_database: BibDatabase,
    exclude: List[str] = [],
    required_fields: Optional[Dict[str, Set[str]]] = None,
) -> None:
    """Check a BibTex database.

    Parameters
    ----------
    bib_database : ``BibDatabase``
        BibTex database.
    %(exclude)s
    %(required_fields)s
    """
    check_type(bib_database, (BibDatabase,), "bib_database")
    check_type(exclude, (list, tuple), "exclude")
    for elt in exclude:
        check_type(elt, (str,))
        check_value(elt, bib_database.entries_dict, "exclude")
    check_type(required_fields, (dict, None), "required_fields")
    if isinstance(required_fields, dict):
        for key, value in required_fields.items():
            check_type(key, (str,))
            check_type(value, (set,))
            for v in value:
                check_type(v, (str,))
    else:
        required_fields, _ = _load_default_config()

    entries = [entry for entry in bib_database.entries if entry["ID"] not in exclude]

    # check for duplicate entries
    _check_duplicate_entries(entries)
    # check minimum fields
    _check_minimum_fields(
        entries,
        required_fields=required_fields,
    )


def _check_duplicate_entries(entries: List[Entry]) -> None:
    """Check for duplicate entries."""
    # check for duplicate entries with the same cite key
    idx = [entry["ID"] for entry in entries]
    if len(idx) != len(set(idx)):
        from collections import Counter

        duplicates = (
            f"{cite_key} ({n})" for cite_key, n in Counter(idx).items() if n != 1
        )
        raise DuplicateEntry(
            "The BibTex file contains duplicate entries with the same cite "
            f"key: {', '.join(duplicates)}."
        )

    # check minimum set of fields for hash
    hash_fields = {"year", "author", "title"}
    for entry in entries:
        if len(hash_fields - set(entry)) != 0:
            raise MissingReqField(
                f"The BibTex file entry '{entry['ID']}' is missing some basic "
                "information: year, author, title."
            )

    # define hash as (title, authors, year)
    hashes = [
        hash((entry["year"], entry["author"], entry["title"])) for entry in entries
    ]
    if len(hashes) != len(set(hashes)):
        from collections import Counter

        duplicates = list()
        duplicate_hashes = [hash_ for hash_, n in Counter(hashes).items() if n != 1]
        hashes = np.array(hashes)
        for hash_ in duplicate_hashes:
            idx = np.where(hashes == hash_)[0]
            duplicates.append(f"({', '.join(entries[k]['ID'] for k in idx)})")
        raise DuplicateEntry(
            "The BibTex file contains duplicate entries with different cite "
            f"keys: {', '.join(duplicates)}."
        )


def _check_minimum_fields(
    entries: List[Entry],
    required_fields: Dict[str, Set[str]],
) -> None:
    """Check that each entry has the minimum required fields."""
    for entry in entries:
        entry_type = entry["ENTRYTYPE"]
        if entry_type not in required_fields:
            continue
        if len(required_fields[entry_type] - set(entry)) != 0:
            raise MissingReqField(
                f"The BibTex file entry '{entry['ID']}' is missing one of "
                f"the required field for a '{entry['ENTRYTYPE']}':"
                f"{', '.join(required_fields[entry_type])}."
            )



class DuplicateEntry(Exception):
    """Error raised when an entry is present multiple times in the same DB."""

    pass


class MissingReqField(Exception):
    """Error raised when a required field is missing from an entry."""

    pass



import logging
from functools import wraps
from pathlib import Path
from typing import Callable, Optional, Union

from ._checks import check_verbose
from ._docs import fill_doc
from ._fixes import _WrapStdOut


@fill_doc
def _init_logger(*, verbose: Optional[Union[bool, str, int]] = None) -> logging.Logger:
    """Initialize a logger.

    Assigns sys.stdout as the first handler of the logger.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    logger : Logger
        The initialized logger.
    """
    # create logger
    verbose = check_verbose(verbose)
    logger = logging.getLogger(__package__.split(".utils", maxsplit=1)[0])
    logger.propagate = False
    logger.setLevel(verbose)

    # add the main handler
    handler = logging.StreamHandler(_WrapStdOut())
    handler.setFormatter(_LoggerFormatter())
    logger.addHandler(handler)

    return logger


@fill_doc
def add_file_handler(
    fname: Union[str, Path],
    mode: str = "a",
    encoding: Optional[str] = None,
    *,
    verbose: Optional[Union[bool, str, int]] = None,
) -> None:
    """Add a file handler to the logger.

    Parameters
    ----------
    fname : str | Path
        Path to the file where the logging output is saved.
    mode : str
        Mode in which the file is opened.
    encoding : str | None
        If not None, encoding used to open the file.
    %(verbose)s
    """
    verbose = check_verbose(verbose)
    handler = logging.FileHandler(fname, mode, encoding)
    handler.setFormatter(_LoggerFormatter())
    handler.setLevel(verbose)
    logger.addHandler(handler)


@fill_doc
def set_log_level(verbose: Union[bool, str, int, None]) -> None:
    """Set the log level for the logger and the first handler ``sys.stdout``.

    Parameters
    ----------
    %(verbose)s
    """
    verbose = check_verbose(verbose)
    logger.setLevel(verbose)


class _LoggerFormatter(logging.Formatter):
    """Format string Syntax."""

    # Format string syntax for the different Log levels
    _formatters = dict()
    _formatters[logging.DEBUG] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s "
        "(%(asctime)s)"
    )
    _formatters[logging.INFO] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s"
    )
    _formatters[logging.WARNING] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s"
    )
    _formatters[logging.ERROR] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s"
    )

    def __init__(self):
        super().__init__(fmt="%(levelname): %(message)s")

    def format(self, record: logging.LogRecord):
        """
        Format the received log record.

        Parameters
        ----------
        record : logging.LogRecord
        """
        if record.levelno <= logging.DEBUG:
            return self._formatters[logging.DEBUG].format(record)
        elif record.levelno <= logging.INFO:
            return self._formatters[logging.INFO].format(record)
        elif record.levelno <= logging.WARNING:
            return self._formatters[logging.WARNING].format(record)
        else:
            return self._formatters[logging.ERROR].format(record)


def verbose(f: Callable) -> Callable:
    """Set the verbose for the function call from the kwargs.

    Parameters
    ----------
    f : callable
        The function with a verbose argument.

    Returns
    -------
    f : callable
        The function.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if "verbose" in kwargs:
            with _use_log_level(kwargs["verbose"]):
                return f(*args, **kwargs)
        else:
            return f(*args, **kwargs)

    return wrapper


@fill_doc
class _use_log_level:
    """Context manager to change the logging level temporary.

    Parameters
    ----------
    %(verbose)s
    """

    def __init__(self, verbose: Union[bool, str, int, None] = None):
        self._old_level = logger.level
        self._level = verbose

    def __enter__(self):
        set_log_level(self._level)

    def __exit__(self, *args):
        set_log_level(self._old_level)


logger = _init_logger(verbose="WARNING")  # equivalent to verbose=None



import argparse

from bibtexparser import dumps

from .. import check_bib_database, clean_bib_database
from .._exception import DuplicateEntry, MissingReqField
from ..config import load_config
from ..io import load_bib, save_bib


class ReturnCode:  # noqa: D101
    no_violations_found = 0
    violations_fixed = 1
    violations_found_unfixable = 2
    invalid_options = 3
    could_not_save = 4


def run():
    """Run bibclean() command."""
    parser = argparse.ArgumentParser(
        prog=f"{__package__.split('.')[0]}", description="cleans a .bib file."
    )
    parser.add_argument(
        "bib",
        type=str,
        metavar="path",
        help="path to the .bib file to clean. If an output is not provided, "
        "this file is overwritten.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        metavar="path",
        help="path to the output .bib file.",
        default=None,
    )
    parser.add_argument(
        "-e",
        "--encoding",
        type=str,
        metavar="str",
        help="encoding of the .bib file.",
        default="utf-8",
    )
    parser.add_argument(
        "--overwrite",
        help="overwrite the file provided in --output if it exists.",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        metavar="path",
        help="path to the TOML configuration.",
        default=None,
    )
    parser.add_argument(
        "--exit-non-zero-on-fix",
        help="exit with a non-zero status code if the .bib file was changed.",
        action="store_true",
    )
    args = parser.parse_args()

    try:
        # load
        bib_database = load_bib(args.bib, encoding=args.encoding)

        # determine configuration and clean
        if args.config is None:
            required_fields = None
            keep_fields = None
            exclude = list()
        else:
            required_fields, keep_fields, exclude = load_config(args.config)
        check_bib_database(bib_database, exclude, required_fields)
        bib_database_clean = clean_bib_database(bib_database, exclude, keep_fields)
    except (DuplicateEntry, MissingReqField):
        return ReturnCode.violations_found_unfixable
    except Exception:
        return ReturnCode.invalid_options

    original = dumps(bib_database)
    cleaned = dumps(bib_database_clean)
    if args.exit_non_zero_on_fix:
        exit_code = (
            ReturnCode.no_violations_found
            if original == cleaned
            else ReturnCode.violations_fixed
        )
    else:
        exit_code = ReturnCode.no_violations_found

    # save
    try:
        output = args.bib if args.output is None else args.output
        overwrite = True if args.output is None else args.overwrite
        save_bib(bib_database, output, encoding=args.encoding, overwrite=overwrite)
    except Exception:
        exit_code = ReturnCode.could_not_save
    return exit_code



from pathlib import Path
from typing import Dict, List, Set, Tuple, Union

from toml import load

from ..utils._docs import fill_doc


@fill_doc
def load_config(
    file: Union[str, Path]
) -> Tuple[Dict[str, Set[str]], Dict[str, Set[str]], List[str]]:
    """Load the ``bibclean`` configuration form a TOML file.

    The loaded configuration is merged with the
    :ref:`default TOML configuration <configuration:default>`, overwriting
    the defaults with the provided configuration.

    Parameters
    ----------
    file : str | Path
        Path to the TOML file containing the ``tool.bibclean`` and
        ``'tool.bibclean.entry_type'`` sections.

    Returns
    -------
    %(required_fields)s
    %(keep_fields)s
    %(exclude)s
    """
    required_fields_def, keep_fields_def = _load_default_config()
    required_fields, keep_fields, exclude, exclude_type = _load_config(file)

    # merge both
    for key, value in required_fields_def.items():
        if key not in required_fields:
            required_fields[key] = required_fields_def[key]
    for key, value in keep_fields_def.items():
        if key not in keep_fields:
            keep_fields[key] = keep_fields_def[key]

    for key, value in required_fields.items():
        if len(value) == 0 and key in required_fields_def:  # absent from TOML
            required_fields[key] = required_fields_def[key]
        elif len(value) == 0 and key not in required_fields_def:
            raise RuntimeError(
                f"The section 'tool.bibclean.{key}' does not have a "
                "'required' default. Please provide the required fields in "
                "your configuration. You can propose a default by creating "
                "an issue on GitHub."
            )
    for key, value in keep_fields.items():
        if len(value) == 0 and key in keep_fields_def:  # absent from TOML
            keep_fields[key] = keep_fields_def[key]
        elif len(value) == 0 and key not in keep_fields_def:
            raise RuntimeError(
                f"The section 'tool.bibclean.{key}' does not have a "
                "'keep' default. Please provide the fields to keep in "
                "your configuration. You can propose a default by creating "
                "an issue on GitHub."
            )

    # exclude by type
    try:
        for exc_type in exclude_type:
            del required_fields[exc_type]
            del keep_fields[exc_type]
    except KeyError:
        pass

    return required_fields, keep_fields, exclude


def _load_config(
    file: Union[str, Path]
) -> Tuple[Dict[str, Set[str]], Dict[str, Set[str]], List[str], List[str]]:
    """Load a configuration from a TOML file."""
    config = load(file)
    if "tool" not in config:
        raise RuntimeError("TOML file is invalid. The section 'tool' is missing.")

    if "bibclean" not in config["tool"]:
        raise RuntimeError(
            "TOML file is invalid. The section(s) 'tool.bibclean' are missing."
        )

    config = config["tool"]["bibclean"]

    required_fields = dict()
    keep_fields = dict()
    exclude = list()
    exclude_type = list()
    for key in config:
        if key == "exclude":
            exclude = config[key]
            for exc in exclude:
                if not isinstance(exc, str):
                    raise TypeError(
                        "TOML file is invalid. The excluded entries must be "
                        "provided as cite-keys in str format. e.g. "
                        "exclude = ['example', 'example_2']."
                    )
            continue
        elif key == "exclude_type":
            exclude_type = config[key]
            for exc in exclude_type:
                if not isinstance(exc, str):
                    raise TypeError(
                        "TOML file is invalid. The excluded types must be "
                        "provided in str format. e.g. "
                        "exclude_type = ['article', 'book']."
                    )
            continue

        # check that TOML is valid
        invalid = sorted(set(config[key].keys()) - {"keep", "required"})
        if len(invalid) != 0:
            plural = "" if len(invalid) == 1 else "s"
            raise KeyError(
                f"TOML file is invalid. Unexpected {invalid} key{plural} in "
                f"section 'bibclean.{key}'."
            )

        try:
            req = config[key]["required"]
        except KeyError:
            req = set()
        try:
            keep = config[key]["keep"]
        except KeyError:
            keep = set()

        if len(req) != len(set(req)) or len(keep) != len(set(keep)):
            raise ValueError(
                "TOML file is invalid. Some fields are present multiple "
                f"times for the same key in section 'bibclean.{key}'."
            )

        required_fields[key] = set(req)
        keep_fields[key] = set(keep)

    return required_fields, keep_fields, exclude, exclude_type


def _load_default_config() -> Tuple[Dict[str, Set[str]], Dict[str, Set[str]]]:
    """Load the default config from 'default.toml'."""
    required_fields, keep_fields, _, _ = _load_config(
        Path(__file__).parent / "default.toml"
    )
    return required_fields, keep_fields



from typing import Dict, List, Optional, Set

from bibtexparser.bibdatabase import BibDatabase

from ._typing import Entry
from .config import _load_default_config
from .utils._checks import check_type, check_value
from .utils._docs import fill_doc
from .utils.logs import logger


@fill_doc
def clean_bib_database(
    bib_database: BibDatabase,
    exclude: List[str] = [],
    keep_fields: Optional[Dict[str, Set[str]]] = None,
) -> BibDatabase:
    """Check and clean a BibTex database.

    Parameters
    ----------
    bib_database : ``BibDatabase``
        BibTex database.
    %(exclude)s
    %(keep_fields)s

    Returns
    -------
    bib_database : ``BibDatabase``
        BibTex database.
    """
    check_type(bib_database, (BibDatabase,), "bib_database")
    check_type(exclude, (list, tuple), "exclude")
    for elt in exclude:
        check_type(elt, (str,))
        check_value(elt, bib_database.entries_dict, "exclude")
    check_type(keep_fields, (dict, None), "keep_fields")
    if isinstance(keep_fields, dict):
        for key, value in keep_fields.items():
            check_type(key, (str,))
            check_type(value, (set,))
            for v in value:
                check_type(v, (str,))
    else:
        _, keep_fields = _load_default_config()

    # reset entries dictionary
    logger.debug("Resetting the entry dictionary.")
    bib_database._entries_dict = {}

    # add URL and DOI to the fields to keep, they will be handled later
    for entry_type in keep_fields:
        keep_fields[entry_type].update({"url", "doi"})

    # remove un-wanted fields
    for k, entry in enumerate(bib_database.entries):
        if entry["ID"] in exclude:
            logger.info("Entry '%s' is listed for exclusion. Skipping.", entry["ID"])
            continue

        entry_type = entry["ENTRYTYPE"]
        if entry_type not in keep_fields:
            continue

        # determine the fields to remove
        fields_to_remove = [
            field for field in set(entry) - keep_fields[entry_type] if field.islower()
        ]
        # logging
        if len(fields_to_remove) == 0:
            logger.debug("No field will be removed from entry '%s'.", entry["ID"])
        else:
            logger.debug(
                "The fields %s will be removed from entry '%s'.",
                fields_to_remove,
                entry["ID"],
            )
        # remove fields
        for field in fields_to_remove:
            del bib_database.entries[k][field]

        # clean doi and url
        need_cleaning = _clean_doi_url(entry)
        if need_cleaning:
            del bib_database.entries[k]["url"]

    # make entries dictionary
    logger.debug("Remaking the entry dictionary.")
    bib_database._make_entries_dict()

    # remove comments
    bib_database.comments = []

    return bib_database


def _clean_doi_url(entry: Entry) -> bool:
    """Look for the fields DOI and URL.

    If only one is present, then we keep it, regardless of which one it is.
    If both are present, we remove the URL.
    """
    field = [key in entry for key in ("url", "doi")]
    if sum(field) == 0:
        logger.warning("Entry '%s' is missing a DOI or URL.", entry["ID"])
    elif sum(field) == 1:
        if field[0]:
            logger.warning("Entry '%s' has a URL instead of a DOI.", entry["ID"])
        return False
    elif sum(field) == 2:
        return True



import argparse
import sys
from copy import deepcopy
from typing import Optional

from bibtexparser import dumps

from .. import check_bib_database, clean_bib_database, logger
from .._exception import DuplicateEntry, MissingReqField
from ..config import load_config
from ..io import load_bib


class ReturnCode:  # noqa: D101
    no_violations_found = 0
    violations_found = 1
    violations_found_unfixable = 2
    invalid_options = 3


def run():
    """Run bibclean-check() command."""
    parser = argparse.ArgumentParser(
        prog=f"{__package__.split('.')[0]}-check",
        description="confirms that a .bib file is processed.",
    )
    parser.add_argument(
        "bib",
        type=str,
        metavar="path",
        help="path to the .bib file to clean. If an output is not provided, "
        "this file is overwritten.",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        metavar="path",
        help="path to the TOML configuration.",
        default=None,
    )
    args = parser.parse_args()

    try:
        sys.exit(_run(args.bib, args.config))
    except KeyboardInterrupt:
        pass


def _run(file: str, config: Optional[str]):
    """Run bibclean-check() command and return an exit-code."""
    try:
        bib_database = load_bib(file)
        if config is None:
            required_fields = None
            keep_fields = None
            exclude = list()
        else:
            required_fields, keep_fields, exclude = load_config(config)
        check_bib_database(bib_database, exclude, required_fields)
        bib_database_clean = clean_bib_database(
            deepcopy(bib_database),
            exclude,
            keep_fields,
        )
    except (DuplicateEntry, MissingReqField):
        return ReturnCode.violations_found_unfixable
    except Exception:
        return ReturnCode.invalid_options

    # check that the file is sorted
    original = dumps(bib_database)
    cleaned = dumps(bib_database_clean)
    if original != cleaned:
        logger.error(
            "The provided '.bib' file is not properly formatted. Please use "
            "the 'bibclean' command to auto-format the entries."
        )
        exit_code = ReturnCode.violations_found
    else:
        exit_code = ReturnCode.no_violations_found
    return exit_code
